import aiohttp
from typing import Any, Literal
from .exceptions import *
from urllib.parse import quote_plus
from io import BytesIO
from json.decoder import JSONDecodeError

class _BaseClient:
    def __init__(self, *, session: aiohttp.ClientSession):
        self._session = session
        self._session_owner = False
        self._URL = "https://api.cobaltonline.net"
        
    @classmethod
    def create(cls):
        session = aiohttp.ClientSession()
        inst = cls(session=session)
        inst._session_owner = True
        return inst
    
    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        return await self.close()
    
    async def _request(self, verb: Literal["GET", "POST", "PUT", "DELETE", "PATCH"], endpoint: str, **kwargs) -> aiohttp.ClientResponse:
        """
        Args:
            verb (str): The HTTP verb to use
            endpoint (str): The URL endpoint to use

        Raises:
            ClosedSessionError: Raises if the HTTP session associated with the client is closed
            ValueError: Raises if the HTTP verb passed is invalid
            ClientRequestError: Raises if an aiohttp error occurs when trying to make the web request

        Returns:
            aiohttp.ClientResponse: The initial response retreived from the server
        """
        if getattr(self, "_session.closed", True):
            raise ClosedSessionError("Cannot operate on a closed/nonexistent client session")
        elif verb not in (valid_verbs := ('GET', 'POST', 'PUT', 'DELETE', 'PATCH')):
            raise ValueError(f"Invalid HTTP verb - expected one of {valid_verbs}, got {verb}")
        else:
            try:
                return await self._session.request(verb, f"{self._URL}{endpoint}", **kwargs)
            except (aiohttp.ClientError, aiohttp.HTTPError) as e:
                raise ClientRequestError(f"Failed to obtain content from endpoint {endpoint}. See the above error for details") from e
     
    async def _get_json_response(self, endpoint: str) -> Any:
        """GET a JSON response from the API

        Args:
            endpoint (str): The endpoint to GET from

        Raises:
            InvalidResponseError: Raises if the client fails to decode the JSON response
        
        Returns:
            Any: The response obtained from the API
        """
        _req = await self._request("GET", endpoint)
        try:
            return (await _req.json())["response"]
        except JSONDecodeError as e:
            raise InvalidResponseError("Failed to decode response body") from e
        
    async def _get_bytes_response(self, endpoint: str, *, bytesio: bool = False) -> bytes | BytesIO:
        """GET a bytes response from the API

        Args:
            endpoint (str): The endpoint to GET from
            bytesio (bool, optional): Whether to wrap the response in a BytesIO buffer. Defaults to False.
            
        Raises:
            InvalidResponseError: Raises if errors are encountered while getting the response bytes

        Returns:
            bytes | BytesIO: The resulting response, in bytes or a BytesIO buffer
        """
        _req = await self._request("GET", endpoint)
        try:
            return await _req.read() if not bytesio else BytesIO(await _req.read())
        except Exception as e:
            raise InvalidResponseError("Failed to process response body") from e
        
    async def close(self):
        """Close the HTTP session associated with the client"""
        if not getattr(self, "_session.closed", True) and self._session_owner:
            return await self._session.close()
        
class Client(_BaseClient):
    
    async def ping(self) -> str:
        """Ping the API server to check if it is online"""
        return await self._get_json_response("/utils/ping")
        
    async def random_fact(self, *, count: int = 1) -> str | list[str]:
        """Get {count} random facts"""
        return await self._get_json_response(f"/random/facts?count={count}")
        
    async def gen_randint(self, floor: int = 0, ceil: int = 25, *, count: int = 1) -> int | list[int]:
        """Generate {count} random integers between {floor} and {ceil}"""
        return await self._get_json_response(f"/random/numbers?floor={floor}&ceil={ceil}&count={count}")
    
    async def image_macro(self, image_url: str, *, top_text: str, bottom_text: str) -> BytesIO:
        """Generate a classic image macro with the Impact font - returns a BytesIO which can be used to directly attach it as a file with your Discord bot"""
        return await self._get_bytes_response(f"/fun/memegen?image_url={image_url}&top_text={quote_plus(top_text)}&bottom_text={quote_plus(top_text)}", bytesio=True)